/*-
 * %%----------------------------------------------------------------------------------------------
 * Solidify Framework - Solidify Frontend - oauth2-interceptor.service.ts
 * SPDX-License-Identifier: GPL-2.0-or-later
 * %----------------------------------------------------------------------------------------------%
 * Copyright (C) 2017 - 2023 University of Geneva
 * %----------------------------------------------------------------------------------------------%
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program.  If not, see
 * <http://www.gnu.org/licenses/gpl-2.0.html>.
 * ----------------------------------------------------------------------------------------------%%
 */

import {
  HttpErrorResponse,
  HttpEvent,
  HttpHandler,
  HttpInterceptor,
  HttpRequest,
  HttpStatusCode,
} from "@angular/common/http";
import {
  Inject,
  Injectable,
} from "@angular/core";
import {Router} from "@angular/router";
import {
  BehaviorSubject,
  filter,
  Observable,
  switchMap,
} from "rxjs";
import {
  catchError,
  mergeMap,
  tap,
} from "rxjs/operators";
import {SOLIDIFY_CONSTANTS} from "../core/constants";
import {DefaultSolidifyEnvironment} from "../core/environments/environment.solidify-defaults";
import {HeaderEnum} from "../core/http/header.enum";
import {ENVIRONMENT} from "../core/injection-tokens/environment.injection-token";
import {AbstractBaseService} from "../core/services/abstract-base.service";
import {
  isArray,
  isFalse,
  isNotNullNorUndefined,
  isNotNullNorUndefinedNorWhiteString,
  isNullOrUndefinedOrWhiteString,
  isTrue,
} from "../core/tools/is/is.tool";
import {SsrUtil} from "../core/utils/ssr.util";

import {EventType} from "./events";
import {OauthHelper} from "./oauth.helper";
import {OAuth2Service} from "./oauth2.service";
import {OAuthStorage} from "./types";

@Injectable({
  providedIn: "root",
})
export class OAuth2Interceptor extends AbstractBaseService implements HttpInterceptor {
  private _errorRefreshTokenBrokeApp: boolean = false;
  private readonly _pendingTokenRefreshedBS: BehaviorSubject<boolean> = new BehaviorSubject<boolean>(false);

  constructor(private readonly authStorage: OAuthStorage,
              private readonly oAuth2Service: OAuth2Service,
              private readonly router: Router,
              @Inject(ENVIRONMENT) private readonly environment: DefaultSolidifyEnvironment,
  ) {
    super();

    this.subscribe(this.oAuth2Service.events.pipe(
      filter(e => e.type === EventType.token_refresh_error),
      tap(e => {
        this._errorRefreshTokenBrokeApp = true;
      }),
    ));
  }

  private _shouldBypassOAuth2TokenHeader(req: HttpRequest<any>): boolean {
    if (req.headers.has(HeaderEnum.skipOauth2TokenHeader) && req.headers.get(HeaderEnum.skipOauth2TokenHeader) === SOLIDIFY_CONSTANTS.STRING_TRUE) {
      return true;
    }
    const url = req.url.toLowerCase();
    if (url === OauthHelper.getTokenEndpoint(this.environment.authorization)) {
      return true;
    }
    if (isArray(this.environment.allowedUrls)) {
      const allowedUrlsToLower = this.environment.allowedUrls.map(a => a.toLowerCase());
      const matchingAllowedUrl = allowedUrlsToLower.find(a => url.indexOf(a) !== -1);
      if (isNotNullNorUndefined(matchingAllowedUrl)) {
        return true;
      }
    }
    if (url.startsWith("assets/") || url.startsWith("./assets/")) {
      return true;
    }
    return false;
  }

  private _removeUndesiredHeader(req: HttpRequest<any>): HttpRequest<any> {
    if (req.headers.has(HeaderEnum.skipOauth2TokenHeader)) {
      const headers = req.headers.delete(HeaderEnum.skipOauth2TokenHeader);
      return req.clone({headers});
    }
    return req;
  }

  public intercept(req: HttpRequest<any>, next: HttpHandler, noRetry: boolean = false): Observable<HttpEvent<any>> {
    if (this._shouldBypassOAuth2TokenHeader(req)) {
      return next.handle(this._removeUndesiredHeader(req));
    }

    const sendAccessToken = this.environment.sendAccessToken;

    if (sendAccessToken) {
      const accessToken = this.oAuth2Service.getAccessToken();
      if (isNotNullNorUndefinedNorWhiteString(accessToken)) {
        const header = "Bearer " + accessToken;
        const headers = req.headers.set("Authorization", header);
        req = req.clone({headers});
      }
    }

    return next
      .handle(req)
      .pipe(
        catchError((err: HttpErrorResponse) => {
          if (isFalse(this._errorRefreshTokenBrokeApp) && !noRetry && (err.status === HttpStatusCode.Unauthorized || err.status === HttpStatusCode.NotAcceptable)) {
            if (isFalse(this.isPendingRefreshTokenRequest)) {
              this._unblockOtherPendingRequestWithInvalidToken();
              const refreshToken = this.oAuth2Service.getRefreshToken();
              if (isNullOrUndefinedOrWhiteString(refreshToken)) {
                this.oAuth2Service.clearStorage();
                this.oAuth2Service.initAuthorizationCodeFlow(this._getRequestedPath());
                throw err;
              }
              return this.oAuth2Service.refreshToken().pipe(
                mergeMap(value => {
                  this._blockOtherRequestWithInvalidToken();
                  if (value) {
                    return this.intercept(req, next, true);
                  }

                  throw err;
                }),
                catchError(err2 => {
                  this._blockOtherRequestWithInvalidToken();
                  this.oAuth2Service.clearStorage();
                  this.oAuth2Service.initAuthorizationCodeFlow(this._getRequestedPath());
                  throw err2;
                }),
              );

            } else {
              return this._pendingTokenRefreshedBS.asObservable().pipe(
                filter(value => isFalse(value)),
                switchMap(() => this.intercept(req, next, this._errorRefreshTokenBrokeApp)),
              );
            }
          }
          throw err;
        }),
      );
  }

  private get isPendingRefreshTokenRequest(): boolean {
    return isTrue(this._pendingTokenRefreshedBS.value);
  }

  private _blockOtherRequestWithInvalidToken(): void {
    this._pendingTokenRefreshedBS.next(false);
  }

  private _unblockOtherPendingRequestWithInvalidToken(): void {
    this._pendingTokenRefreshedBS.next(true);
  }

  private _getRequestedPath(): string {
    const baseHref = this.environment.baseHref;
    let requestedUrl = SsrUtil.window?.location.pathname;
    if (isNotNullNorUndefinedNorWhiteString(baseHref) && baseHref !== SOLIDIFY_CONSTANTS.URL_SEPARATOR && requestedUrl.startsWith(baseHref)) {
      requestedUrl = requestedUrl.substring(baseHref.length);
      requestedUrl = SOLIDIFY_CONSTANTS.URL_SEPARATOR + requestedUrl;
    }
    if (requestedUrl === SOLIDIFY_CONSTANTS.URL_SEPARATOR) {
      return "";
    }
    return requestedUrl;
  }
}
